"""
Generic forces and energies
"""
import pandas as pd
import numpy as np

from ..utils import to_nd
from . import units

from .planar_gradients import area_grad as area_grad2d
from .planar_gradients import lumen_area_grad
from .sheet_gradients import height_grad, area_grad
from .bulk_gradients import volume_grad, lumen_volume_grad


def elastic_force(element_df, var, elasticity, prefered):
    params = {"x": var, "K": elasticity, "x0": prefered}
    force = element_df.eval("{K} * ({x} - {x0})".format(**params))
    return force


def _elastic_force(element_df, x, elasticity, prefered):
    force = element_df[elasticity] * (element_df[x] - element_df[prefered])
    return force


def elastic_energy(element_df, var, elasticity, prefered):
    params = {"x": var, "K": elasticity, "x0": prefered}
    energy = element_df.eval("0.5 * {K} * ({x} - {x0}) ** 2".format(**params))
    return energy


def _elastic_energy(element_df, x, elasticity, prefered):
    energy = 0.5 * element_df[elasticity] * (element_df[x] - element_df[prefered]) ** 2
    return np.array(energy)


class AbstractEffector:
    """ The effector class is used by model factories
    to construct a model.


    """

    dimensions = None
    magnitude = None
    spatial_ref = None, None
    temporal_ref = None, None

    label = "Abstract effector"
    element = None  # cell, face, edge or vert
    specs = {"cell": {}, "face": {}, "edge": {}, "vert": {}}

    @staticmethod
    def energy(eptm):
        raise NotImplementedError

    @staticmethod
    def gradient(eptm):
        raise NotImplementedError

    @staticmethod
    def get_nrj_norm(specs):
        raise NotImplementedError


#     @classmethod
#     @property
#     def __doc__(cls):
#         f"""Effector implementing {cls.label} with a magnitude factor
#  {cls.magnitude}.

# Works on an `Epithelium` object's {cls.element} elements.
# """


class LengthElasticity(AbstractEffector):
    """Elastic half edge
    """

    dimensions = units.line_elasticity
    label = "Length elasticity"
    magnitude = "length_elasticity"
    element = "edge"
    spatial_ref = "prefered_length", units.length

    specs = {
        "edge": {
            "is_active": 1,
            "length": 1.0,
            "length_elasticity": 1.0,
            "prefered_length": 1.0,
            "ux": (1 / 3) ** 0.5,
            "uy": (1 / 3) ** 0.5,
            "uz": (1 / 3) ** 0.5,
        }
    }

    @staticmethod
    def get_nrj_norm(specs):
        return (
            specs["edge"]["length_elasticity"] * specs["edge"]["prefered_length"] ** 2
        )

    @staticmethod
    def energy(eptm):
        return elastic_energy(
            eptm.edge_df, "length", "length_elasticity * is_active", "prefered_length"
        )

    @staticmethod
    def gradient(eptm):
        kl_l0 = elastic_force(
            eptm.edge_df, "length", "length_elasticity * is_active", "prefered_length"
        )
        grad = eptm.edge_df[eptm.ucoords] * to_nd(kl_l0, eptm.dim)
        grad.columns = ["g" + u for u in eptm.coords]
        return -grad, grad


class FaceAreaElasticity(AbstractEffector):

    dimensionless = False
    dimensions = units.area_elasticity
    magnitude = "area_elasticity"
    label = "Area elasticity"
    element = "face"
    specs = {
        "face": {
            "is_alive": 1,
            "area": 1.0,
            "area_elasticity": 1.0,
            "prefered_area": 1.0,
        },
        "edge": {"sub_area": 1 / 6.0},
    }

    spatial_ref = "prefered_area", units.area

    @staticmethod
    def get_nrj_norm(specs):
        return specs["face"]["area_elasticity"] * specs["face"]["prefered_area"] ** 2

    @staticmethod
    def energy(eptm):
        return elastic_energy(
            eptm.face_df, "area", "area_elasticity * is_alive", "prefered_area"
        )

    @staticmethod
    def gradient(eptm):
        ka_a0_ = elastic_force(
            eptm.face_df, "area", "area_elasticity * is_alive", "prefered_area"
        )
        ka_a0 = to_nd(eptm.upcast_face(ka_a0_), len(eptm.coords))

        if len(eptm.coords) == 2:
            grad_a_srce, grad_a_trgt = area_grad2d(eptm)
        elif len(eptm.coords) == 3:
            grad_a_srce, grad_a_trgt = area_grad(eptm)

        grad_a_srce = ka_a0 * grad_a_srce
        grad_a_trgt = ka_a0 * grad_a_trgt

        grad_a_srce.columns = ["g" + u for u in eptm.coords]
        grad_a_trgt.columns = ["g" + u for u in eptm.coords]

        return grad_a_srce, grad_a_trgt


class FaceVolumeElasticity(AbstractEffector):

    dimensions = units.vol_elasticity
    magnitude = "vol_elasticity"
    label = "Volume elasticity"
    element = "face"
    specs = {
        "face": {"is_alive": 1, "vol": 1.0, "vol_elasticity": 1.0, "prefered_vol": 1.0},
        "vert": {"height": 1.0},
        "edge": {"sub_area": 1 / 6},
    }

    spatial_ref = "prefered_vol", units.vol

    @staticmethod
    def get_nrj_norm(specs):
        return specs["face"]["vol_elasticity"] * specs["face"]["prefered_vol"] ** 2

    @staticmethod
    def energy(eptm):
        return elastic_energy(
            eptm.face_df, "vol", "vol_elasticity * is_alive", "prefered_vol"
        )

    @staticmethod
    def gradient(eptm):
        kv_v0_ = elastic_force(
            eptm.face_df, "vol", "vol_elasticity * is_alive", "prefered_vol"
        )

        kv_v0 = to_nd(eptm.upcast_face(kv_v0_), 3)

        edge_h = to_nd(eptm.upcast_srce(eptm.vert_df["height"]), 3)
        area_ = eptm.edge_df["sub_area"]
        area = to_nd(area_, 3)
        grad_a_srce, grad_a_trgt = area_grad(eptm)
        grad_h = eptm.upcast_srce(height_grad(eptm))

        grad_v_srce = kv_v0 * (edge_h * grad_a_srce + area * grad_h)
        grad_v_trgt = kv_v0 * (edge_h * grad_a_trgt)

        grad_v_srce.columns = ["g" + u for u in eptm.coords]
        grad_v_trgt.columns = ["g" + u for u in eptm.coords]

        return grad_v_srce, grad_v_trgt


class CellAreaElasticity(AbstractEffector):

    dimensions = units.area_elasticity
    magnitude = "area_elasticity"
    label = "Area elasticity"
    element = "cell"
    specs = {
        "cell": {
            "is_alive": 1,
            "area": 1.0,
            "area_elasticity": 1.0,
            "prefered_area": 1.0,
        }
    }
    spatial_ref = "prefered_area", units.area

    @staticmethod
    def get_nrj_norm(specs):
        return specs["cell"]["area_elasticity"] * specs["cell"]["prefered_area"] ** 2

    @staticmethod
    def energy(eptm):
        return elastic_energy(eptm.cell_df, "area", "area_elasticity", "prefered_area")

    @staticmethod
    def gradient(eptm):
        ka_a0_ = elastic_force(
            eptm.cell_df, "area", "area_elasticity * is_alive", "prefered_area"
        )

        ka_a0 = to_nd(eptm.upcast_cell(ka_a0_), 3)

        grad_a_srce, grad_a_trgt = area_grad(eptm)

        grad_a_srce = ka_a0 * grad_a_srce
        grad_a_trgt = ka_a0 * grad_a_trgt
        grad_a_srce.columns = ["g" + u for u in eptm.coords]
        grad_a_trgt.columns = ["g" + u for u in eptm.coords]

        return grad_a_srce, grad_a_trgt


class CellVolumeElasticity(AbstractEffector):

    dimensions = units.vol_elasticity
    magnitude = "vol_elasticity"
    label = "Volume elasticity"
    element = "cell"
    spatial_ref = "prefered_vol", units.vol

    specs = {
        "cell": {"is_alive": 1, "vol": 1.0, "vol_elasticity": 1.0, "prefered_vol": 1.0}
    }

    @staticmethod
    def get_nrj_norm(specs):
        return specs["cell"]["vol_elasticity"] * specs["cell"]["prefered_vol"] ** 2

    @staticmethod
    def energy(eptm):
        return elastic_energy(eptm.cell_df, "vol", "vol_elasticity", "prefered_vol")

    @staticmethod
    def gradient(eptm):
        kv_v0_ = elastic_force(
            eptm.cell_df, "vol", "vol_elasticity * is_alive", "prefered_vol"
        )

        kv_v0 = to_nd(eptm.upcast_cell(kv_v0_), 3)
        grad_v_srce, grad_v_trgt = volume_grad(eptm)
        grad_v_srce = kv_v0 * grad_v_srce
        grad_v_trgt = kv_v0 * grad_v_trgt

        grad_v_srce.columns = ["g" + u for u in eptm.coords]
        grad_v_trgt.columns = ["g" + u for u in eptm.coords]

        return grad_v_srce, grad_v_trgt


class LumenVolumeElasticity(AbstractEffector):
    """
    Global volume elasticity of the object.
    For example the volume of the yolk in the Drosophila embryo
    """

    dimensions = units.vol_elasticity
    magnitude = "lumen_vol_elasticity"
    label = "Lumen volume elasticity"
    element = "settings"
    spatial_ref = "lumen_prefered_vol", units.vol

    specs = {
        "settings": {
            "lumen_vol": 1.0,
            "lumen_vol_elasticity": 1.0,
            "lumen_prefered_vol": 1.0,
        }
    }

    @staticmethod
    def get_nrj_norm(specs):
        return (
            specs["settings"]["lumen_vol_elasticity"]
            * specs["settings"]["lumen_prefered_vol"] ** 2
        )

    @staticmethod
    def energy(eptm):

        return _elastic_energy(
            eptm.settings, "lumen_vol", "lumen_vol_elasticity", "lumen_prefered_vol"
        )

    @staticmethod
    def gradient(eptm):
        kv_v0 = _elastic_force(
            eptm.settings, "lumen_vol", "lumen_vol_elasticity", "lumen_prefered_vol"
        )

        grad_v_srce, grad_v_trgt = lumen_volume_grad(eptm)
        grad_v_srce = kv_v0 * grad_v_srce
        grad_v_trgt = kv_v0 * grad_v_trgt

        grad_v_srce.columns = ["g" + u for u in eptm.coords]
        grad_v_trgt.columns = ["g" + u for u in eptm.coords]

        return grad_v_srce, grad_v_trgt


class LineTension(AbstractEffector):

    dimensions = units.line_tension
    magnitude = "line_tension"
    label = "Line tension"
    element = "edge"
    specs = {"edge": {"is_active": 1, "line_tension": 1.0}}

    spatial_ref = "mean_length", units.length

    @staticmethod
    def energy(eptm):
        return eptm.edge_df.eval(
            "line_tension" "* is_active" "* length / 2"
        )  # accounts for half edges

    @staticmethod
    def gradient(eptm):
        grad_srce = -eptm.edge_df[eptm.ucoords] * to_nd(
            eptm.edge_df.eval("line_tension * is_active/2"), len(eptm.coords)
        )
        grad_srce.columns = ["g" + u for u in eptm.coords]
        grad_trgt = -grad_srce
        return grad_srce, grad_trgt


class FaceContractility(AbstractEffector):

    dimensions = units.line_elasticity
    magnitude = "contractility"
    label = "Contractility"
    element = "face"
    specs = {"face": {"is_alive": 1, "perimeter": 1.0, "contractility": 1.0}}

    spatial_ref = "mean_perimeter", units.length

    @staticmethod
    def energy(eptm):
        return eptm.face_df.eval("0.5 * is_alive * contractility * perimeter ** 2")

    @staticmethod
    def gradient(eptm):

        gamma_ = eptm.face_df.eval("contractility * perimeter * is_alive")
        gamma = eptm.upcast_face(gamma_)

        grad_srce = -eptm.edge_df[eptm.ucoords] * to_nd(gamma, len(eptm.coords))
        grad_srce.columns = ["g" + u for u in eptm.coords]
        grad_trgt = -grad_srce
        return grad_srce, grad_trgt


class SurfaceTension(AbstractEffector):

    dimensions = units.area_tension
    magnitude = "surface_tension"

    spatial_ref = "prefered_area", units.area

    label = "Surface tension"
    element = "face"
    specs = {"face": {"is_active": 1, "surface_tension": 1.0, "area": 1.0}}

    @staticmethod
    def energy(eptm):

        return eptm.face_df.eval("surface_tension * area")

    @staticmethod
    def gradient(eptm):

        G = to_nd(eptm.upcast_face(eptm.face_df["surface_tension"]), len(eptm.coords))
        grad_a_srce, grad_a_trgt = area_grad(eptm)

        grad_a_srce = G * grad_a_srce
        grad_a_trgt = G * grad_a_trgt
        grad_a_srce.columns = ["g" + u for u in eptm.coords]
        grad_a_trgt.columns = ["g" + u for u in eptm.coords]

        return grad_a_srce, grad_a_trgt


class LineViscosity(AbstractEffector):

    dimensions = units.line_viscosity
    magnitude = "edge_viscosity"

    label = "Linear viscosity"
    element = "edge"
    spatial_ref = "mean_length", units.length
    temporal_ref = "dt", units.time
    specs = {"edge": {"is_active": 1, "edge_viscosity": 1.0}}

    @staticmethod
    def gradient(eptm):
        grad_srce = eptm.edge_df[["vx", "vy", "vz"]] * to_nd(
            eptm.edge_df["edge_viscosity"], len(eptm.coords)
        )
        grad_srce.columns = ["g" + u for u in eptm.coords]
        return grad_srce, None


class BorderElasticity(AbstractEffector):
    dimensions = units.line_elasticity
    label = "Border edges elasticity"
    magnitude = "border_elasticity"
    element = "edge"
    spatial_ref = "prefered_length", units.length

    specs = {
        "edge": {
            "is_active": 1,
            "length": 1.0,
            "border_elasticity": 1.0,
            "prefered_length": 1.0,
            "is_border": 1.0,
        }
    }

    @staticmethod
    def get_nrj_norm(specs):
        return (
            specs["edge"]["border_elasticity"] * specs["edge"]["prefered_length"] ** 2
        )

    @staticmethod
    def energy(eptm):
        return elastic_energy(
            eptm.edge_df,
            "length",
            "border_elasticity * is_active * is_border / 2",
            "prefered_length",
        )

    @staticmethod
    def gradient(eptm):

        kl_l0 = elastic_force(
            eptm.edge_df,
            var="length",
            elasticity="border_elasticity * is_active * is_border",
            prefered="prefered_length",
        )
        grad = eptm.edge_df[eptm.ucoords] * to_nd(kl_l0, eptm.dim)
        grad.columns = ["g" + u for u in eptm.coords]
        return grad / 2, -grad / 2


class LumenAreaElasticity(AbstractEffector):
    """

    .. math:: \frac{K_Y}{2}(A_{\mathrm{lumen}} - A_{0,\mathrm{lumen}})^2

    """

    dimensions = units.area_elasticity
    label = "Lumen volume constraint"
    magnitude = "lumen_elasticity"
    element = "settings"
    spatial_ref = "lumen_prefered_vol", units.area

    specs = {
        "settings": {
            "lumen_elasticity": 1.0,
            "lumen_prefered_vol": 1.0,
            "lumen_vol": 1.0,
        }
    }

    @staticmethod
    def energy(eptm):
        Ky = eptm.settings["lumen_elasticity"]
        V0 = eptm.settings["lumen_prefered_vol"]
        Vy = eptm.settings["lumen_vol"]
        return np.array([Ky * (Vy - V0) ** 2 / 2])

    @staticmethod
    def gradient(eptm):
        Ky = eptm.settings["lumen_elasticity"]
        V0 = eptm.settings["lumen_prefered_vol"]
        Vy = eptm.settings["lumen_vol"]
        grad_srce, grad_trgt = lumen_area_grad(eptm)
        return (Ky * (Vy - V0) * grad_srce, Ky * (Vy - V0) * grad_trgt)


def _exponants(dimensions, ref_dimensions, spatial_unit=None, temporal_unit=None):

    spatial_exponant = time_exponant = 0
    rel_dimensionality = (dimensions / ref_dimensions).dimensionality

    if spatial_unit is not None:
        spatial_exponant = (
            rel_dimensionality.get(units.length, 0)
            / spatial_unit.dimensionality[units.length]
        )

    if temporal_unit is not None:
        time_exponant = (
            rel_dimensionality.get(units.time, 0)
            / temporal_unit.dimensionality[units.time]
        )
    return spatial_exponant, time_exponant


def scaler(nondim_specs, dim_specs, effector, ref_effector):
    spatial_val, spatial_unit = ref_effector.spatial_ref
    temporal_val, temporal_unit = ref_effector.temporal_ref

    s_expo, t_expo = _exponants(
        effector.dimensions, ref_effector.dimensions, spatial_unit, temporal_unit
    )

    ref_magnitude = ref_effector.magnitude
    ref_element = ref_effector.element
    factor = (
        dim_specs[ref_element][ref_magnitude]
        * dim_specs[ref_element].get(spatial_val, 1) ** s_expo
        * dim_specs[ref_element].get(temporal_val, 1) ** t_expo
    )
    return factor


def dimensionalize(nondim_specs, dim_specs, effector, ref_effector):
    magnitude = effector.magnitude
    element = effector.element
    factor = scaler(nondim_specs, dim_specs, effector, ref_effector)
    dim_specs[element][magnitude] = factor * nondim_specs[element][magnitude]
    return dim_specs


def normalize(dim_specs, nondim_specs, effector, ref_effector):
    magnitude = effector.magnitude
    element = effector.element
    factor = scaler(nondim_specs, dim_specs, effector, ref_effector)
    dim_specs[element][magnitude] = nondim_specs[element][magnitude] / factor
    return dim_specs



import numpy as np


def update_spherical(eptm):
    """ Computes the spherical coordinates (rho, theta, phi)
    of an epithelium.

    rho is the distance to the coordinate system's origin, theta
    is the co-latitude (0 ≤ θ < π) and phi is the longitude (0 ≤ ϕ < 2π).
    """
    for element in ["vert", "face", "cell"]:
        if eptm.datasets.get(element) is None:
            continue
        eptm.datasets[element]["rho"] = np.linalg.norm(
            eptm.datasets[element][["x", "y", "z"]], axis=1
        )
        eptm.datasets[element]["theta"] = np.arccos(
            eptm.datasets[element].eval("z / rho")
        )
        eptm.datasets[element]["phi"] = np.arctan2(
            eptm.datasets[element].eval("y / (rho * sin(theta))"),
            eptm.datasets[element].eval("x / (rho * sin(theta))"),
        )


def rotation_matrix(angle, direction):
    # Copyright (c) 2006-2015, Christoph Gohlke
    # Copyright (c) 2006-2015, The Regents of the University of California
    # Produced at the Laboratory for Fluorescence Dynamics
    # All rights reserved.
    #
    # Redistribution and use in source and binary forms, with or without
    # modification, are permitted provided that the following conditions
    #  are met:
    #
    # * Redistributions of source code must retain the above copyright
    #   notice, this list of conditions and the following disclaimer.
    # * Redistributions in binary form must reproduce the above copyright
    #   notice, this list of conditions and the following disclaimer in the
    #   documentation and/or other materials provided with the distribution.
    # * Neither the name of the copyright holders nor the names of any
    #   contributors may be used to endorse or promote products derived
    #   from this software without specific prior written permission.
    #
    # THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
    # "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
    # LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
    # FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
    # IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
    # LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
    # CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
    # SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
    # INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
    # CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
    # ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
    # OF THE POSSIBILITY OF SUCH DAMAGE.
    """
    Returns the 3X3 rotation matrix around `direction` by `angle`

    adapted from http://www.lfd.uci.edu/~gohlke/code/transformations.py.html
    """
    sina = np.sin(angle)
    cosa = np.cos(angle)
    direction = direction / np.linalg.norm(direction)
    # rotation matrix around unit vector
    R = np.diag([cosa, cosa, cosa])
    R += np.outer(direction, direction) * (1.0 - cosa)
    direction *= sina
    R += np.array(
        [
            [0.0, -direction[2], direction[1]],
            [direction[2], 0.0, -direction[0]],
            [-direction[1], direction[0], 0.0],
        ]
    )
    return R



import logging
import numpy as np
from functools import wraps

from .base_topology import add_vert
from tyssue.utils.decorators import do_undo, validate


logger = logging.getLogger(name=__name__)


def type1_transition(sheet, edge01, epsilon=0.1, remove_tri_faces=True):
    """Performs a type 1 transition around the edge edge01

    See ../../doc/illus/t1_transition.png for a sketch of the definition
    of the vertices and cells letterings

    Parameters
    ----------
    sheet : a `Sheet` instance
    edge_01 : int
       index of the edge around which the transition takes place
    epsilon : float, optional
       default 0.1, the initial length of the new edge.
    remove_tri_faces : bool, optional
       if True (the default), will remove triangular cells
       after the T1 transition is performed
    """
    # Grab the neighbours
    vert0, vert1, face_b = sheet.edge_df.loc[edge01, ["srce", "trgt", "face"]].astype(
        int
    )
    if sheet.face_df.loc[face_b, "num_sides"] < 4:
        logger.warning(
            """Face %s has 3 sides,
type 1 transition is not allowed"""
            % face_b
        )
        return face_b

    edge10_ = sheet.edge_df[
        (sheet.edge_df["srce"] == vert1) & (sheet.edge_df["trgt"] == vert0)
    ]
    if not len(edge10_.index):
        # edge01 is at a border, we need to add the opposite
        # half edge
        edge10_ = sheet.edge_df.loc[edge01].copy()
        sheet.edge_df = sheet.edge_df.append(edge10_, ignore_index=True)
        edge10 = sheet.edge_df.index[-1]
        sheet.edge_df.loc[edge10, ["srce", "trgt", "face"]] = vert1, vert0, -1
        face_d = -1
    else:
        edge10 = edge10_.index
        face_d = _cast_to_int(edge10_["face"])

    if face_d != -1 and sheet.face_df.loc[face_d, "num_sides"] < 4:
        logger.warning(
            """Face %s has 3 sides,
        type 1 transition is not allowed"""
            % face_d
        )
        return face_d
    edge20_ = sheet.edge_df[
        (sheet.edge_df["trgt"] == vert0) & (sheet.edge_df["face"] == face_b)
    ]
    edge20 = edge20_.index
    vert2 = _cast_to_int(edge20_["srce"])

    edge02_ = sheet.edge_df[
        (sheet.edge_df["srce"] == vert0) & (sheet.edge_df["trgt"] == vert2)
    ]
    edge02 = edge02_.index
    face_a = _cast_to_int(edge02_["face"])

    edge13_ = sheet.edge_df[
        (sheet.edge_df["srce"] == vert1) & (sheet.edge_df["face"] == face_b)
    ]
    edge13 = edge13_.index
    vert3 = _cast_to_int(edge13_["trgt"])

    edge31_ = sheet.edge_df[
        (sheet.edge_df["srce"] == vert3) & (sheet.edge_df["trgt"] == vert1)
    ]
    edge31 = edge31_.index
    face_c = _cast_to_int(edge31_["face"])

    edge13_ = sheet.edge_df[
        (sheet.edge_df["srce"] == vert1) & (sheet.edge_df["face"] == face_b)
    ]
    edge13 = edge13_.index
    vert3 = _cast_to_int(edge13_["trgt"])
    if face_a != -1:
        edge50_ = sheet.edge_df[
            (sheet.edge_df["trgt"] == vert0) & (sheet.edge_df["face"] == face_a)
        ]
        edge50 = edge50_.index
        vert5 = _cast_to_int(edge50_["srce"])

        edge05_ = sheet.edge_df[
            (sheet.edge_df["srce"] == vert0) & (sheet.edge_df["trgt"] == vert5)
        ]
        edge05 = edge05_.index

    elif face_d != -1:
        edge05_ = sheet.edge_df[
            (sheet.edge_df["srce"] == vert0) & (sheet.edge_df["face"] == face_d)
        ]
        edge05 = edge05_.index
        vert5 = _cast_to_int(edge05_["trgt"])

        edge50_ = sheet.edge_df[
            (sheet.edge_df["srce"] == vert5) & (sheet.edge_df["trgt"] == vert0)
        ]
        edge50 = edge50_.index

    else:
        raise ValueError("Edge has no neighbour around vertex %d", vert0)

    logger.debug("faces a, b, c, d")
    logger.debug("%d, %d, %d, %d", face_a, face_b, face_c, face_d)
    logger.debug("vertices 0, 1, 2, 3, 5")
    logger.debug("%d, %d, %d, %d; %d", vert0, vert1, vert2, vert3, vert5)

    # Perform the rearangements
    sheet.edge_df.loc[edge01, "face"] = face_c
    sheet.edge_df.loc[edge10, "face"] = face_a
    sheet.edge_df.loc[edge13, ["srce", "trgt", "face"]] = vert0, vert3, face_b
    sheet.edge_df.loc[edge31, ["srce", "trgt", "face"]] = vert3, vert0, face_c

    sheet.edge_df.loc[edge50, ["srce", "trgt", "face"]] = vert5, vert1, face_a
    sheet.edge_df.loc[edge05, ["srce", "trgt", "face"]] = vert1, vert5, face_d

    # Displace the vertices
    mean_pos = (
        sheet.vert_df.loc[vert0, sheet.coords] + sheet.vert_df.loc[vert1, sheet.coords]
    ) / 2
    face_b_pos = sheet.face_df.loc[face_b, sheet.coords]
    sheet.vert_df.loc[vert0, sheet.coords] = (
        mean_pos - (mean_pos - face_b_pos) * epsilon
    )
    if face_d != -1:
        face_d_pos = sheet.face_df.loc[face_d, sheet.coords]
    else:
        face_d_pos = mean_pos

    sheet.vert_df.loc[vert1, sheet.coords] = (
        mean_pos - (mean_pos - face_d_pos) * epsilon
    )

    sheet.edge_df = sheet.edge_df[sheet.edge_df["face"] != -1].copy()
    sheet.edge_df.index.name = "edge"
    sheet.reset_topo()
    if not remove_tri_faces:
        return 0
    # Type 1 transitions might create 3 or 2 sided cells, we remove those
    tri_faces = sheet.face_df[sheet.face_df["num_sides"] < 4].index
    while len(tri_faces):
        remove_face(sheet, tri_faces[0])
        tri_faces = sheet.face_df[sheet.face_df["num_sides"] < 4].index

    return 0


def cell_division(sheet, mother, geom, angle=None):

    if not sheet.face_df.loc[mother, "is_alive"]:
        logger.warning("Cell %s is not alive and cannot devide", mother)
        return
    edge_a, edge_b = get_division_edges(sheet, mother, geom, angle=angle, axis="x")
    if edge_a is None:
        return

    vert_a, new_edge_a, new_opp_edge_a = add_vert(sheet, edge_a)
    vert_b, new_edge_b, new_opp_edge_b = add_vert(sheet, edge_b)
    sheet.vert_df.index.name = "vert"
    daughter = face_division(sheet, mother, vert_a, vert_b)
    return daughter


def get_division_edges(sheet, mother, geom, angle=None, axis="x"):

    if angle is None:
        angle = np.random.random() * np.pi

    m_data = sheet.edge_df[sheet.edge_df["face"] == mother]
    # if angle == 0:
    #     face_pos = sheet.face_df.loc[mother, sheet.coords]
    #     rot_pos = sheet.vert_df[sheet.coords].copy()
    #     for c in sheet.coords:
    #         rot_pos.loc[:, c] = rot_pos[c] - face_pos[c]
    # else:
    rot_pos = geom.face_projected_pos(sheet, mother, psi=angle)

    srce_pos = rot_pos.loc[m_data["srce"], axis]
    srce_pos.index = m_data.index
    trgt_pos = rot_pos.loc[m_data["trgt"], axis]
    trgt_pos.index = m_data.index
    try:
        edge_a = m_data[(srce_pos < 0) & (trgt_pos >= 0)].index[0]
        edge_b = m_data[(srce_pos >= 0) & (trgt_pos < 0)].index[0]
    except IndexError:
        print("Failed")
        logger.error("Division of Cell {} failed".format(mother))
        return None, None
    return edge_a, edge_b


def face_division(sheet, mother, vert_a, vert_b):
    """
    Divides the face associated with edges
    indexed by `edge_a` and `edge_b`, splitting it
    in the middle of those edes.
    """
    # mother = sheet.edge_df.loc[edge_a, 'face']

    face_cols = sheet.face_df.loc[mother]
    sheet.face_df = sheet.face_df.append(face_cols, ignore_index=True)
    sheet.face_df.index.name = "face"
    daughter = int(sheet.face_df.index[-1])

    edge_cols = sheet.edge_df[sheet.edge_df["face"] == mother].iloc[0]
    sheet.edge_df = sheet.edge_df.append(edge_cols, ignore_index=True)
    new_edge_m = sheet.edge_df.index[-1]
    sheet.edge_df.loc[new_edge_m, "srce"] = vert_b
    sheet.edge_df.loc[new_edge_m, "trgt"] = vert_a

    sheet.edge_df = sheet.edge_df.append(edge_cols, ignore_index=True)
    new_edge_d = sheet.edge_df.index[-1]
    sheet.edge_df.loc[new_edge_d, "srce"] = vert_a
    sheet.edge_df.loc[new_edge_d, "trgt"] = vert_b

    # ## Discover daughter edges
    m_data = sheet.edge_df[sheet.edge_df["face"] == mother]
    daughter_edges = [new_edge_d]
    srce, trgt = vert_a, vert_b
    srces, trgts = m_data[["srce", "trgt"]].values.T

    while trgt != vert_a:
        srce, trgt = trgt, trgts[srces == trgt][0]
        daughter_edges.append(
            m_data[(m_data["srce"] == srce) & (m_data["trgt"] == trgt)].index[0]
        )
    sheet.edge_df.loc[daughter_edges, "face"] = daughter
    sheet.edge_df.index.name = "edge"
    sheet.reset_topo()
    return daughter


def remove_face(sheet, face):

    if np.isnan(sheet.face_df.loc[face, "num_sides"]):
        logger.info("Face %i is not valid, aborting")
        return

    edges = sheet.edge_df[sheet.edge_df["face"] == face]
    verts = edges["srce"].values

    out_orbits = sheet.get_orbits("srce", "trgt")
    in_orbits = sheet.get_orbits("trgt", "srce")

    new_vert_data = sheet.vert_df.loc[verts].mean()
    sheet.vert_df = sheet.vert_df.append(new_vert_data, ignore_index=True)
    new_vert = sheet.vert_df.index[-1]
    for v in verts:
        out_jes = out_orbits.loc[v].index
        sheet.edge_df.loc[out_jes, "srce"] = new_vert

        in_jes = in_orbits.loc[v].index
        sheet.edge_df.loc[in_jes, "trgt"] = new_vert

    sheet.edge_df = sheet.edge_df[sheet.edge_df["srce"] != sheet.edge_df["trgt"]]

    sheet.edge_df = sheet.edge_df[sheet.edge_df["face"] != face].copy()

    fidx = sheet.face_df.index.delete(face)
    sheet.face_df = sheet.face_df.loc[fidx].copy()

    logger.info("removed {} of {} vertices ".format(len(verts), sheet.vert_df.shape[0]))
    logger.info("face {} is now dead ".format(face))

    vidx = sheet.vert_df.index.delete(verts)
    sheet.vert_df = sheet.vert_df.loc[vidx].copy()
    sheet.reset_index()
    sheet.reset_topo()
    return new_vert


def split_vert(sheet, vert, epsilon=0.0):
    """
    Splits (or opens up) the sheet at vertex `vert`, creating
    new verts and separating the connected opposite edges, see
    ../../doc/illus/vertex_split.png

    Parameters
    ----------
    sheet: a :class:`Sheet` instance
    vert: int, index of the vertex to split
    epsilon: float, the relative amount of recoil
      of the new vertices towards the face centers
    """
    # Grab relevant edges
    vert_out_edges = sheet.edge_df[(sheet.edge_df["srce"] == vert)]
    vert_in_edges = sheet.edge_df[(sheet.edge_df["trgt"] == vert)]
    # Grab relevant faces
    neighbor_faces = set(vert_out_edges["face"])
    if len(neighbor_faces) == 1:
        logger.info(
            """
Chosen vertex %i is bound to a single cell, nothing to do"""
            % vert
        )
        return

    # Create the new vertices
    num_new_vert = len(neighbor_faces) - 1
    vert_data = sheet.vert_df.loc[[vert] * num_new_vert]
    sheet.vert_df = sheet.vert_df.append(vert_data, ignore_index=True)

    new_verts = list(sheet.vert_df.index[-num_new_vert:])
    split_verts = [vert] + new_verts

    # reasign sources and targets in edge_df
    for f, v in zip(neighbor_faces, split_verts):
        eo = vert_out_edges[vert_out_edges["face"] == f].index
        sheet.edge_df.loc[eo, "srce"] = v
        ei = vert_in_edges[vert_in_edges["face"] == f].index
        sheet.edge_df.loc[ei, "trgt"] = v
    sheet.reset_topo()
    sheet.reset_index()

    faces = sheet.face_df.loc[neighbor_faces]
    verts = sheet.vert_df.loc[split_verts]
    dr = -verts[sheet.coords] + faces[sheet.coords].values
    sheet.vert_df.loc[split_verts, sheet.coords] += dr * epsilon


def resolve_t1s(sheet, geom, model, solver, max_iter=60):

    l_th = sheet.settings["threshold_length"]
    i = 0
    while sheet.edge_df.length.min() < l_th:

        for edge in (
            sheet.edge_df[sheet.edge_df.length < l_th].sort_values("length").index
        ):
            try:
                type1_transition(sheet, edge)
            except KeyError:
                continue
            sheet.reset_index()
            sheet.reset_topo()
            geom.update_all(sheet)
        solver.find_energy_min(sheet, geom, model)
        i += 1
        if i > max_iter:
            break


def _cast_to_int(df_value):

    if len(df_value) == 1:
        return int(df_value)
    elif len(df_value) == 0:
        return -1
    else:
        raise ValueError(
            "Trying to retrieve an integer from " "a more than length 1 df "
        )



import numpy as np


class BaseGeometry:
    """
    """

    @staticmethod
    def update_all(sheet):
        raise NotImplementedError

    @staticmethod
    def scale(sheet, delta, coords):
        """ Scales the coordinates `coords`
        by a factor `delta`
        """
        sheet.vert_df[coords] = sheet.vert_df[coords] * delta

    @staticmethod
    def update_dcoords(sheet):
        """
        Update the edge vector coordinates  on the
        `coords` basis (`default_coords` by default).
        Modifies the corresponding
        columns (i.e `['dx', 'dy', 'dz']`) in sheet.edge_df.

        Also updates the upcasted coordinates of the source and target
        vertices
        """
        if sheet.settings.get("boundaries") is None:
            data = sheet.vert_df[sheet.coords]
            srce_pos = sheet.upcast_srce(data)
            trgt_pos = sheet.upcast_trgt(data)
            sheet.edge_df[["s" + c for c in sheet.coords]] = srce_pos
            sheet.edge_df[["t" + c for c in sheet.coords]] = trgt_pos
            sheet.edge_df[sheet.dcoords] = trgt_pos - srce_pos
        else:
            update_periodic_dcoords(sheet)

    @staticmethod
    def update_length(sheet):
        """
        Updates the edge_df `length` column on the `coords` basis
        """
        sheet.edge_df["length"] = np.linalg.norm(sheet.edge_df[sheet.dcoords], axis=1)

    @staticmethod
    def update_perimeters(sheet):
        """
        Updates the perimeter of each face.
        """
        sheet.face_df["perimeter"] = sheet.sum_face(sheet.edge_df["length"])

    @staticmethod
    def update_centroid(sheet):
        """
        Updates the face_df `coords` columns as the face's vertices
        center of mass. Also updates the edge_df fx, fy, fz columns
        with their upcasted values
        """
        scoords = ["s" + c for c in sheet.coords]
        sheet.face_df[sheet.coords] = sheet.edge_df.groupby("face")[scoords].mean()
        face_pos = sheet.upcast_face(sheet.face_df[sheet.coords])
        sheet.edge_df[["f" + c for c in sheet.coords]] = face_pos

    @staticmethod
    def center(eptm):
        """
        Transates the epithelium vertices so that the center
        of mass is at the center of the coordinate system,
        and updates the geometry
        """

        eptm.vert_df[eptm.coords] = (
            eptm.vert_df[eptm.coords].values
            - eptm.vert_df[eptm.coords].mean(axis=0).values[np.newaxis, :]
        )

    @staticmethod
    def dist_to_point(vert_df, point, coords):
        """
        Returns the distance of all vertices from point over the
        coordinates

        Parameters
        ----------
        vert_df: a :class:`pandas.DataFrame` with the points coordinates
          in the columns given by the `coords` argument

        point: a doublet (in 2D) or triplet (in 3D) giving the reference point
          coordinates

        coords: list of 2 or 3 strings giving the column names

        Returns
        -------

        distance: a :class:`pandas.Series` with the same length
          as the input `vert_df`
        """
        return sum(((vert_df[c] - u) ** 2 for c, u in zip(coords, point))) ** 0.5


def update_periodic_dcoords(sheet):
    """ Updates the coordinates for periodic boundary conditions.

    """
    for u, boundary in sheet.settings["boundaries"].items():
        period = boundary[1] - boundary[0]
        shift = period * (
            -(sheet.vert_df[u] > boundary[1]).astype(float)
            + (sheet.vert_df[u] <= boundary[0]).astype(float)
        )
        sheet.vert_df[u] = sheet.vert_df[u] + shift

    srce_pos = sheet.upcast_srce(sheet.vert_df[sheet.coords])
    trgt_pos = sheet.upcast_trgt(sheet.vert_df[sheet.coords])
    sheet.edge_df[["s" + u for u in sheet.coords]] = srce_pos
    sheet.edge_df[["t" + u for u in sheet.coords]] = trgt_pos
    sheet.edge_df[sheet.dcoords] = trgt_pos - srce_pos
    for u, boundary in sheet.settings["boundaries"].items():
        period = boundary[1] - boundary[0]
        center = boundary[1] - period / 2
        shift = period * (
            -(sheet.edge_df["d" + u] >= period / 2).astype(float)
            + (sheet.edge_df["d" + u] < -period / 2).astype(float)
        )
        sheet.edge_df["d" + u] = sheet.edge_df["d" + u] + shift
        sheet.edge_df[f"at_{u}_boundary"] = shift != 0

        sheet.face_df[f"at_{u}_boundary"] = sheet.edge_df.groupby("face")[
            f"at_{u}_boundary"
        ].apply(any)
        f_at_boundary = sheet.upcast_face(sheet.face_df[f"at_{u}_boundary"]).astype(int)
        period = boundary[1] - boundary[0]
        srce_shift = f_at_boundary * (sheet.edge_df["s" + u] < center) * period
        trgt_shift = f_at_boundary * (sheet.edge_df["t" + u] < center) * period
        sheet.edge_df["s" + u] += srce_shift
        sheet.edge_df["t" + u] += trgt_shift



"""
Hexagonal grids
---------------
"""
from math import tau

import numpy as np
import pandas as pd

from ..config.geometry import flat_sheet
from .utils import make_df


def hexa_grid2d(nx, ny, distx, disty, noise=None):
    """Creates an hexagonal grid of points
    """
    cy, cx = np.mgrid[0:ny, 0:nx]
    cx = cx.astype(np.float)
    cy = cy.astype(np.float)
    cx[::2, :] += 0.5

    centers = np.vstack([cx.flatten(), cy.flatten()]).astype(np.float).T
    centers[:, 0] *= distx
    centers[:, 1] *= disty
    if noise is not None:
        pos_noise = np.random.normal(scale=noise, size=centers.shape)
        centers += pos_noise
    return centers


def hexa_grid3d(nx, ny, nz, distx=1.0, disty=1.0, distz=1.0, noise=None):
    """Creates an hexagonal grid of points
    """
    cz, cy, cx = np.mgrid[0:nz, 0:ny, 0:nx]
    cx = cx.astype(np.float)
    cy = cy.astype(np.float)
    cz = cz.astype(np.float)
    cx[:, ::2] += 0.5
    cy[::2, :] += 0.5
    cy *= np.sqrt(3) / 2
    cz *= np.sqrt(3) / 2

    centers = np.vstack([cx.flatten(), cy.flatten(), cz.flatten()]).T
    centers[:, 0] *= distx
    centers[:, 1] *= disty
    centers[:, 2] *= distz
    if noise is not None:
        pos_noise = np.random.normal(scale=noise, size=centers.shape)
        centers += pos_noise
    return centers


def circle(num_t, radius=1.0, phase=0.0):
    """Returns x and y positions of `num_t` points regularly placed around a circle
    of radius `radius`, shifted by `phase` radians.

    Parameters
    ----------
    num_t : int
        the number of points around the circle
    radius : float, default 1.
        the radius of the circle
    phase : float, default 0.0
        angle shift w/r to the x axis in radians

    Returns
    -------
    points : np.Ndarray of shape (num_t, 2), the x, y positions of the points

    """
    if not num_t:
        return np.zeros((1, 2))

    theta = np.arange(0, tau, tau / num_t)
    return np.vstack([radius * np.cos(theta + phase), radius * np.sin(theta + phase)]).T


def hexa_disk(num_t, radius=1):
    """Returns an arrays of x, y positions of points evenly spread on
    a disk with num_t points on the periphery.

    Parameters
    ----------
    num_t : int
        the number of poitns on the disk periphery, the rest of the disk is
        filled automaticaly
    radius : float, default 1.
        the radius of the disk

    """

    n_circles = int(np.ceil(num_t / tau) + 1)
    if not n_circles:
        return np.zeros((1, 2))

    num_ts = np.linspace(num_t, 0, n_circles, dtype=int)
    rads = radius * num_ts / num_t
    phases = np.pi * num_ts / num_t
    return np.concatenate(
        [circle(n, r, phi) for n, r, phi in zip(num_ts, rads, phases)]
    )


def hexa_cylinder(num_t, num_z, radius=1.0, capped=False, noise=0):
    """Returns an arrays of x, y positions of points evenly spread on
    a cylinder with num_t points on the periphery and num_z points on its length.

    Parameters
    ----------
    num_t : int,
        The number of points on the periphery
    num_z : int,
        The number of points along the z axis (the length of the cylinder)
    radius : float, default 1
        The radius of the cylinder
    capped : bool, default False
        If True, the tips of the cylinder are capped by a disk of point
        as generated by the `hexa_disk` function.
    noise : float, default 0
        normaly distributed position noise around the cell points

    """
    delta_t = tau / num_t
    points_zt = hexa_grid2d(num_z, num_t, delta_t, delta_t, noise=noise)
    points_zt[:, 0] = points_zt[:, 0] * radius

    zs = points_zt[:, 0]
    xs = radius * np.cos(points_zt[:, 1])
    ys = radius * np.sin(points_zt[:, 1])
    points_xyz = np.vstack((xs - xs.mean(), ys - ys.mean(), zs - zs.mean())).T

    if capped:

        disk = hexa_disk(num_t, radius)[num_t:, :]
        left_cap = np.zeros((disk.shape[0], 3))
        left_cap[:, :2] = disk
        left_cap[:, 2] = points_xyz[:, 2].min()  # - delta_t  # * 0.5

        right_cap = np.zeros((disk.shape[0], 3))
        right_cap[:, :2] = disk
        right_cap[:, 2] = points_xyz[:, 2].max()  # + delta_t  # * 0.5
        points_xyz = np.concatenate([left_cap, points_xyz, right_cap])

    return points_xyz


"""

Three cells sheet
-----------------
"""


def three_faces_sheet_array():
    """
    Creates the apical junctions mesh of three packed hexagonal faces.
    If `zaxis` is `True` (defaults to False), adds a `z` coordinates,
    with `z = 0`.

    Faces have a side length of 1.0 +/- 1e-3.

    Returns
    -------

    points: (13, ndim) np.array of floats
      the positions, where ndim is 2 or 3 depending on `zaxis`
    edges: (15, 2)  np.array of ints
      indices of the edges
    (Nc, Nv, Ne): triple of ints
      number of faces, vertices and edges (3, 13, 15)

    """
    Nc = 3  # Number of faces
    points = np.array(
        [
            [0.0, 0.0],
            [1.0, 0.0],
            [1.5, 0.866],
            [1.0, 1.732],
            [0.0, 1.732],
            [-0.5, 0.866],
            [-1.5, 0.866],
            [-2, 0.0],
            [-1.5, -0.866],
            [-0.5, -0.866],
            [0, -1.732],
            [1, -1.732],
            [1.5, -0.866],
        ]
    )

    edges = np.array(
        [
            [0, 1],
            [1, 2],
            [2, 3],
            [3, 4],
            [4, 5],
            [5, 0],
            [5, 6],
            [6, 7],
            [7, 8],
            [8, 9],
            [9, 0],
            [9, 10],
            [10, 11],
            [11, 12],
            [12, 1],
        ]
    )

    Nv, Ne = len(points), len(edges)
    return points, edges, (Nc, Nv, Ne)


def three_faces_sheet(zaxis=False):
    """
    Creates the apical junctions mesh of three packed hexagonal faces.
    If `zaxis` is `True` (defaults to False), adds a `z` coordinates,
    with `z = 0`.

    Faces have a side length of 1.0 +/- 1e-3.

    Returns
    -------

    face_df: the faces `DataFrame` indexed from 0 to 2
    vert_df: the junction vertices `DataFrame`
    edge_df: the junction edges `DataFrame`

    """
    points, _, (Nc, Nv, Ne) = three_faces_sheet_array()

    if zaxis:
        coords = ["x", "y", "z"]
    else:
        coords = ["x", "y"]

    face_idx = pd.Index(range(Nc), name="face")
    vert_idx = pd.Index(range(Nv), name="vert")

    _edge_e_idx = np.array(
        [
            [0, 1, 0],
            [1, 2, 0],
            [2, 3, 0],
            [3, 4, 0],
            [4, 5, 0],
            [5, 0, 0],
            [0, 5, 1],
            [5, 6, 1],
            [6, 7, 1],
            [7, 8, 1],
            [8, 9, 1],
            [9, 0, 1],
            [0, 9, 2],
            [9, 10, 2],
            [10, 11, 2],
            [11, 12, 2],
            [12, 1, 2],
            [1, 0, 2],
        ]
    )

    edge_idx = pd.Index(range(_edge_e_idx.shape[0]), name="edge")

    specifications = flat_sheet()

    # ## Faces DataFrame
    face_df = make_df(index=face_idx, spec=specifications["face"])

    # ## Junction vertices and edges DataFrames
    vert_df = make_df(index=vert_idx, spec=specifications["vert"])
    edge_df = make_df(index=edge_idx, spec=specifications["edge"])

    edge_df["srce"] = _edge_e_idx[:, 0]
    edge_df["trgt"] = _edge_e_idx[:, 1]
    edge_df["face"] = _edge_e_idx[:, 2]

    vert_df.loc[:, coords[:2]] = points
    if zaxis:
        vert_df.loc[:, coords[2:]] = 0.0

    datasets = {"face": face_df, "vert": vert_df, "edge": edge_df}
    return datasets, specifications



from . import effectors
from .factory import model_factory

PlanarModel = model_factory(
    [effectors.LineTension, effectors.FaceContractility, effectors.FaceAreaElasticity],
    effectors.FaceAreaElasticity,
)



"""
Energy minimization solvers for the sheet vertex model
"""
from scipy import optimize
from .. import config
import numpy as np


class Solver:
    @classmethod
    def find_energy_min(cls, sheet, geom, model, pos_idx=None, **settings_kw):
        # Loads 'tyssue/config/solvers/minimize.json
        settings = config.solvers.minimize_spec()
        settings.update(**settings_kw)

        coords = sheet.coords
        if pos_idx is None:
            pos_idx = sheet.vert_df[sheet.vert_df["is_active"] == 1].index

        pos0 = sheet.vert_df.loc[pos_idx, coords].values.ravel()

        max_length = 2 * sheet.edge_df["length"].max()
        bounds = np.vstack([pos0 - max_length, pos0 + max_length]).T
        res = optimize.minimize(
            cls.opt_energy,
            pos0,
            args=(pos_idx, sheet, geom, model),
            bounds=bounds,
            jac=cls.opt_grad,
            **settings["minimize"]
        )
        return res

    @staticmethod
    def set_pos(pos, pos_idx, sheet):
        ndims = len(sheet.coords)
        pos_ = pos.reshape((pos.size // ndims, ndims))
        sheet.vert_df.loc[pos_idx, sheet.coords] = pos_

    @classmethod
    def opt_energy(cls, pos, pos_idx, sheet, geom, model):
        cls.set_pos(pos, pos_idx, sheet)
        geom.update_all(sheet)
        return model.compute_energy(sheet, full_output=False)

    # The unused arguments bellow are legit, need same call sig as above
    @staticmethod
    def opt_grad(pos, pos_idx, sheet, geom, model):
        grad_i = model.compute_gradient(sheet, components=False)
        grad_i = grad_i.loc[pos_idx]
        return grad_i.values.flatten()

    @classmethod
    def approx_grad(cls, sheet, geom, model):
        pos0 = sheet.vert_df[sheet.coords].values.ravel()
        pos_idx = sheet.vert_df[sheet.vert_df["is_active"] == 1].index

        grad = optimize.approx_fprime(
            pos0, cls.opt_energy, 1e-9, pos_idx, sheet, geom, model
        )
        return grad

    @classmethod
    def check_grad(cls, sheet, geom, model):

        pos_idx = sheet.vert_df[sheet.vert_df["is_active"] == 1].index
        pos0 = sheet.vert_df.loc[pos_idx, sheet.coords].values.ravel()

        grad_err = optimize.check_grad(
            cls.opt_energy, cls.opt_grad, pos0.flatten(), pos_idx, sheet, geom, model
        )
        return grad_err



import logging

from .core import objects
from .version import version as __version__
from .core.objects import Epithelium
from .core.sheet import Sheet
from .core.monolayer import Monolayer, MonolayerWithLamina
from .core.multisheet import MultiSheet
from .core.history import History
from .geometry.planar_geometry import PlanarGeometry
from .geometry.sheet_geometry import SheetGeometry
from .geometry.bulk_geometry import (
    BulkGeometry,
    MonolayerGeometry,
    RNRGeometry,
    ClosedMonolayerGeometry,
)

from .geometry.multisheetgeometry import MultiSheetGeometry

logger = logging.getLogger('tyssue')



"""
Matplotlib based plotting
"""
import warnings
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

from matplotlib import cm
from matplotlib.path import Path
from matplotlib.patches import Polygon, FancyArrow, Arc, PathPatch
from matplotlib.collections import PatchCollection, PolyCollection
from ..config.draw import sheet_spec
from ..utils.utils import spec_updater, get_sub_eptm

COORDS = ["x", "y"]


def sheet_view(sheet, coords=COORDS, ax=None, **draw_specs_kw):
    """ Base view function, parametrizable
    through draw_secs

    The default sheet_spec specification is:

    {'edge': {
      'visible': True,
      'width': 0.5,
      'head_width': 0.2, # arrow head width for the edges
      'length_includes_head': True, # see matplotlib Arrow artist doc
      'shape': 'right',
      'color': '#2b5d0a', # can be an array
      'alpha': 0.8,
      'zorder': 1,
      'colormap': 'viridis'},
     'vert': {
      'visible': True,
      's': 100,
      'color': '#000a4b',
      'alpha': 0.3,
      'zorder': 2},
     'face': {
      'visible': False,
      'color': '#8aa678',
      'alpha': 1.0,
      'zorder': -1}
      }
    """
    draw_specs = sheet_spec()
    spec_updater(draw_specs, draw_specs_kw)
    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.get_figure()

    vert_spec = draw_specs["vert"]
    if vert_spec["visible"]:
        ax = draw_vert(sheet, coords, ax, **vert_spec)

    edge_spec = draw_specs["edge"]
    if edge_spec["visible"]:
        ax = draw_edge(sheet, coords, ax, **edge_spec)

    face_spec = draw_specs["face"]
    if face_spec["visible"]:
        ax = draw_face(sheet, coords, ax, **face_spec)

    ax.autoscale()
    ax.set_aspect("equal")
    ax.grid()
    return fig, ax


def draw_face(sheet, coords, ax, **draw_spec_kw):
    """Draws epithelial sheet polygonal faces in matplotlib
    Keyword values can be specified at the element
    level as columns of the sheet.face_df
    """

    draw_spec = sheet_spec()["face"]
    draw_spec.update(**draw_spec_kw)
    collection_specs = parse_face_specs(draw_spec, sheet)

    if "visible" in sheet.face_df.columns:
        edges = sheet.edge_df[sheet.upcast_face(sheet.face_df["visible"])].index
        _sheet = get_sub_eptm(sheet, edges)
        if _sheet is not None:
            sheet = _sheet
            color = collection_specs["facecolors"]
            if isinstance(color, np.ndarray):
                faces = sheet.face_df["face_o"].values.astype(np.uint32)
                collection_specs["facecolors"] = color.take(faces, axis=0)

    polys = sheet.face_polygons(coords)
    p = PolyCollection(polys, closed=True, **collection_specs)
    ax.add_collection(p)
    return ax


def parse_face_specs(face_draw_specs, sheet):

    collection_specs = {}
    color = face_draw_specs.get("color")
    if color is None:
        return {}
    elif isinstance(color, str):
        collection_specs["facecolors"] = color
    elif hasattr(color, "__len__"):
        collection_specs["facecolors"] = _face_color_from_sequence(
            face_draw_specs, sheet
        )
    if "alpha" in face_draw_specs:
        collection_specs["alpha"] = face_draw_specs["alpha"]

    return collection_specs


def _face_color_from_sequence(face_spec, sheet):
    color_ = face_spec["color"]
    cmap = cm.get_cmap(face_spec.get("colormap", "viridis"))
    color_min, color_max = face_spec.get("color_range", (color_.min(), color_.max()))

    if color_.shape in [(sheet.Nf, 3), (sheet.Nf, 4)]:
        return color_

    elif color_.shape == (sheet.Nf,):
        if np.ptp(color_) < 1e-10:
            warnings.warn("Attempting to draw a colormap " "with a uniform value")
            return np.ones((sheet.Nf, 3)) * 0.5

        normed = (color_ - color_min) / (color_max - color_min)
        return cmap(normed)

    else:
        raise ValueError(
            "shape of `face_spec['color']` must be either (Nf, 3), (Nf, 4) or (Nf,)"
        )


def draw_vert(sheet, coords, ax, **draw_spec_kw):
    """Draw junction vertices in matplotlib
    """
    draw_spec = sheet_spec()["vert"]
    draw_spec.update(**draw_spec_kw)

    x, y = coords
    if "z_coord" in sheet.vert_df.columns:
        pos = sheet.vert_df.sort_values("z_coord")[coords]
    else:
        pos = sheet.vert_df[coords]
    ax.scatter(pos[x], pos[y], **draw_spec_kw)
    return ax


def draw_edge(sheet, coords, ax, **draw_spec_kw):
    """
    """
    draw_spec = sheet_spec()["edge"]
    draw_spec.update(**draw_spec_kw)

    x, y = coords
    dx, dy = ("d" + c for c in coords)
    app_length = np.hypot(sheet.edge_df[dx], sheet.edge_df[dy])

    patches = []
    arrow_specs, collections_specs = parse_edge_specs(draw_spec, sheet)

    for idx, edge in sheet.edge_df[app_length > 1e-6].iterrows():
        srce = int(edge["srce"])
        arrow = FancyArrow(
            sheet.vert_df.loc[srce, x],
            sheet.vert_df.loc[srce, y],
            sheet.edge_df.loc[idx, dx],
            sheet.edge_df.loc[idx, dy],
            **arrow_specs
        )
        patches.append(arrow)
    ax.add_collection(PatchCollection(patches, False, **collections_specs))
    return ax


def parse_edge_specs(edge_draw_specs, sheet):

    arrow_keys = ["head_width", "length_includes_head", "shape"]
    arrow_specs = {
        key: val for key, val in edge_draw_specs.items() if key in arrow_keys
    }
    collection_specs = {}
    if "color" in edge_draw_specs:
        if isinstance(edge_draw_specs["color"], str):
            collection_specs["edgecolors"] = edge_draw_specs["color"]
        elif hasattr(edge_draw_specs["color"], "__len__"):
            collection_specs["edgecolors"] = _wire_color_from_sequence(
                edge_draw_specs, sheet
            )

    if "width" in edge_draw_specs:
        collection_specs["linewidths"] = edge_draw_specs["width"]
    if "alpha" in edge_draw_specs:
        collection_specs["alpha"] = edge_draw_specs["alpha"]
    return arrow_specs, collection_specs


def _wire_color_from_sequence(edge_spec, sheet):
    """
    """
    color_ = edge_spec["color"]
    color_min, color_max = edge_spec.get("color_range", (color_.min(), color_.max()))
    cmap = cm.get_cmap(edge_spec.get("colormap", "viridis"))
    if color_.shape in [(sheet.Nv, 3), (sheet.Nv, 4)]:
        return (sheet.upcast_srce(color_) + sheet.upcast_trgt(color_)) / 2
    elif color_.shape == (sheet.Nv,):
        if np.ptp(color_) < 1e-10:
            warnings.warn("Attempting to draw a colormap " "with a uniform value")
            return np.ones((sheet.Ne, 3)) * 0.7
        if not hasattr(color_, "index"):
            color_ = pd.Series(color_, index=sheet.vert_df.index)
        color_ = (sheet.upcast_srce(color_) + sheet.upcast_trgt(color_)) / 2
        return cmap((color_ - color_min) / (color_max - color_min))

    elif color_.shape in [(sheet.Ne, 3), (sheet.Ne, 4)]:
        return color_
    elif color_.shape == (sheet.Ne,):
        if np.ptp(color_) < 1e-10:
            warnings.warn("Attempting to draw a colormap " "with a uniform value")
            return np.ones((sheet.Nv, 3)) * 0.7
        return cmap((color_ - color_min) / (color_max - color_min))


def quick_edge_draw(sheet, coords=["x", "y"], ax=None, **draw_spec_kw):

    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.get_figure()

    x, y = coords
    srce_x = sheet.upcast_srce(sheet.vert_df[x]).values
    srce_y = sheet.upcast_srce(sheet.vert_df[y]).values
    trgt_x = sheet.upcast_trgt(sheet.vert_df[x]).values
    trgt_y = sheet.upcast_trgt(sheet.vert_df[y]).values

    lines_x, lines_y = np.zeros(2 * sheet.Ne), np.zeros(2 * sheet.Ne)
    lines_x[::2] = srce_x
    lines_x[1::2] = trgt_x
    lines_y[::2] = srce_y
    lines_y[1::2] = trgt_y
    # Trick from https://github.com/matplotlib/
    # matplotlib/blob/master/lib/matplotlib/tri/triplot.py#L65
    lines_x = np.insert(lines_x, slice(None, None, 2), np.nan)
    lines_y = np.insert(lines_y, slice(None, None, 2), np.nan)
    ax.plot(lines_x, lines_y, **draw_spec_kw)
    ax.set_aspect("equal")
    return fig, ax


def plot_forces(
    sheet, geom, model, coords, scaling, ax=None, approx_grad=None, **draw_specs_kw
):
    """Plot the net forces at each vertex, with their amplitudes multiplied
    by `scaling`. To be clear, this is the oposite of the gradient - grad E.
    """
    draw_specs = sheet_spec()
    spec_updater(draw_specs, draw_specs_kw)
    gcoords = ["g" + c for c in coords]
    if approx_grad is not None:
        app_grad = approx_grad(sheet, geom, model)
        grad_i = (
            pd.DataFrame(
                index=sheet.active_verts,
                data=app_grad.reshape((-1, len(sheet.coords))),
                columns=["g" + c for c in sheet.coords],
            )
            * scaling
        )

    else:
        grad_i = model.compute_gradient(sheet, components=False) * scaling

    arrows = pd.DataFrame(columns=coords + gcoords, index=sheet.vert_df.index)
    arrows[coords] = sheet.vert_df[coords]
    arrows[gcoords] = -grad_i[gcoords]  # F = -grad E

    if ax is None:
        fig, ax = quick_edge_draw(sheet, coords)
    else:
        fig = ax.get_figure()

    for _, arrow in arrows.iterrows():
        ax.arrow(*arrow, **draw_specs["grad"])
    return fig, ax


def plot_scaled_energies(sheet, geom, model, scales, ax=None):
    """Plot scaled energies

    Parameters
    ----------
    sheet: a:class: Sheet object
    geom: a :class:`Geometry` class
    model: a :class:'Model'
    scales: np.linspace of float

    Returns
    -------
    fig: a :class:matplotlib.figure.Figure instance
    ax: :class:matplotlib.Axes instance, default None
    """

    from ..utils import scaled_unscaled

    def get_energies():
        energies = np.array([e.mean() for e in model.compute_energy(sheet, True)])

        return energies

    energies = np.array(
        [scaled_unscaled(get_energies, scale, sheet, geom) for scale in scales]
    )
    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.get_figure()
    ax.plot(scales, energies.sum(axis=1), "k-", lw=4, alpha=0.3, label="total")
    for e, label in zip(energies.T, model.labels):
        ax.plot(scales, e, label=label)
    ax.legend()
    return fig, ax


def plot_analytical_to_numeric_comp(sheet, model, geom, isotropic_model, nondim_specs):

    iso = isotropic_model
    fig, ax = plt.subplots(figsize=(8, 8))

    deltas = np.linspace(0.1, 1.8, 50)

    lbda = nondim_specs["edge"]["line_tension"]
    gamma = nondim_specs["face"]["contractility"]

    ax.plot(
        deltas,
        iso.isotropic_energy(deltas, nondim_specs),
        "k-",
        label="Analytical total",
    )
    try:
        ax.plot(sheet.delta_o, iso.isotropic_energy(sheet.delta_o, nondim_specs), "ro")
    except:
        pass
    ax.plot(deltas, iso.elasticity(deltas), "b-", label="Analytical volume elasticity")
    ax.plot(
        deltas,
        iso.contractility(deltas, gamma),
        color="orange",
        ls="-",
        label="Analytical contractility",
    )
    ax.plot(deltas, iso.tension(deltas, lbda), "g-", label="Analytical line tension")

    ax.set_xlabel(r"Isotropic scaling $\delta$")
    ax.set_ylabel(r"Isotropic energie $\bar E$")

    energies = iso.isotropic_energies(sheet, model, geom, deltas, nondim_specs)
    # energies = energies / norm
    ax.plot(
        deltas,
        energies[:, 2],
        "bo:",
        lw=2,
        alpha=0.8,
        label="Computed volume elasticity",
    )
    ax.plot(
        deltas, energies[:, 0], "go:", lw=2, alpha=0.8, label="Computed line tension"
    )
    ax.plot(
        deltas,
        energies[:, 1],
        ls=":",
        marker="o",
        color="orange",
        label="Computed contractility",
    )
    ax.plot(
        deltas, energies.sum(axis=1), "ko:", lw=2, alpha=0.8, label="Computed total"
    )

    ax.legend(loc="upper left", fontsize=10)
    ax.set_ylim(0, 1.2)
    print(sheet.delta_o, deltas[energies.sum(axis=1).argmin()])

    return fig, ax


def get_arc_data(sheet):

    srce_pos = sheet.upcast_srce(sheet.vert_df[sheet.coords])
    trgt_pos = sheet.upcast_trgt(sheet.vert_df[sheet.coords])

    radius = 1 / sheet.edge_df["curvature"]

    e_x = sheet.edge_df["dx"] / sheet.edge_df["length"]
    e_y = sheet.edge_df["dy"] / sheet.edge_df["length"]

    center_x = (srce_pos.x + trgt_pos.x) / 2 - e_y * (radius - sheet.edge_df["sagitta"])

    center_y = (srce_pos.y + trgt_pos.y) / 2 - e_x * (radius - sheet.edge_df["sagitta"])

    alpha = sheet.edge_df["arc_chord_angle"]
    beta = sheet.edge_df["chord_orient"]

    # Ok, I admit a fair amount of trial and
    # error to get to the stuff below :-p
    rot = beta - np.sign(alpha) * np.pi / 2
    theta1 = (-alpha + rot) * np.sign(alpha)
    theta2 = (alpha + rot) * np.sign(alpha)

    center_data = pd.DataFrame.from_dict(
        {
            "radius": np.abs(radius),
            "x": center_x,
            "y": center_y,
            "theta1": theta1,
            "theta2": theta2,
        }
    )
    return center_data


def curved_view(sheet, radius_cutoff=1e3):

    center_data = get_arc_data(sheet)
    fig, ax = sheet_view(sheet, **{"edge": {"visible": False}})

    curves = []
    for idx, edge in center_data.iterrows():
        if edge["radius"] > radius_cutoff:
            st = sheet.edge_df.loc[idx, ["srce", "trgt"]]
            xy = sheet.vert_df.loc[st, sheet.coords]
            patch = PathPatch(Path(xy))
        else:
            patch = Arc(
                edge[["x", "y"]],
                2 * edge["radius"],
                2 * edge["radius"],
                theta1=edge["theta1"] * 180 / np.pi,
                theta2=edge["theta2"] * 180 / np.pi,
            )
        curves.append(patch)
    ax.add_collection(PatchCollection(curves, False, **{"facecolors": "none"}))
    ax.autoscale()
    return fig, ax



"""
An epithelial sheet, i.e a 2D mesh in a 2D or 3D space,
akin to a HalfEdge data structure in CGAL.

For purely 2D the geometric properties are defined in
 `tyssue.geometry.planar_geometry`
A dynamical model derived from Fahradifar et al. 2007 is provided in
`tyssue.dynamics.planar_vertex_model`


For 2D in 3D, the geometric properties are defined in
 `tyssue.geometry.sheet_geometry`
A dynamical model derived from Monier, Gettings et al. 2015 is provided in
`tyssue.dynamics.sheet_vertex_model`


"""

import warnings
import numpy as np
import pandas as pd

from .objects import Epithelium
from ..config.geometry import flat_sheet


class Sheet(Epithelium):
    """
    An epithelial sheet, i.e a 2D mesh in a 2D or 3D space,
    akin to a HalfEdge data structure in CGAL.

    The geometric properties are defined in `tyssue.geometry.sheet_geometry`
    A dynamical model derived from Fahradifar et al. 2007 is provided in
    `tyssue.dynamics.sheet_vertex_model`


    """

    def __init__(self, identifier, datasets, specs=None, coords=None):
        """
        Creates an epithelium sheet, such as the apical junction network.

        Parameters
        ----------
        identifier: `str`, the tissue name
        face_df: `pandas.DataFrame` indexed by the faces indexes
            this df holds the vertices associated with

        """
        if specs is None:
            specs = flat_sheet()
        super().__init__(identifier, datasets, specs, coords)

    def reset_topo(self):
        super().reset_topo()
        if "opposite" in self.edge_df.columns:
            self.edge_df["opposite"] = get_opposite(self.edge_df)

    def get_opposite(self):
        self.edge_df["opposite"] = get_opposite(self.edge_df)

    def get_neighbors(self, face):
        """Returns the faces adjacent to `face`
        """
        return super().get_neighbors(face, elem="face")

    def get_neighborhood(self, face, order):
        """Returns `face` neighborhood up to a degree of `order`

        For example, if `order` is 2, it wil return the adjacent, faces
        and theses faces neighbors.

        Returns
        -------
        neighbors : pd.DataFrame with two colums, the index
            of the neighboring face, and it's neighboring order

        """
        # Start with the face so that it's not gathered later
        return super().get_neighborhood(face, order, elem="face")

    def get_extra_indices(self):
        """Computes extra indices:

        - `self.free_edges`: half-edges at the epithelium boundary
        - `self.dble_edges`: half-edges inside the epithelium,
          with an opposite
        - `self.east_edges`: half of the `dble_edges`, pointing east
          (figuratively)
        - `self.west_edges`: half of the `dble_edges`, pointing west
           (the order of the east and west edges is conserved, so that
           the ith west half-edge is the opposite of the ith east half-edge)
        - `self.sgle_edges`: joint index over free and east edges, spanning
           the entire graph without double edges
        - `self.wrpd_edges`: joint index over free edges followed by the
           east edges twice, such that a vector over the whole half-edge
            dataframe is wrapped over the single edges
        - `self.srtd_edges`: index over the whole half-edge sorted such that
           the free edges come first, then the east, then the west

        Also computes:
        - `self.Ni`: the number of inside full edges
          (i.e. `len(self.east_edges)`)
        - `self.No`: the number of outside full edges
          (i.e. `len(self.free_edges)`)
        - `self.Nd`: the number of double half edges
          (i.e. `len(self.dble_edges)`)
        - `self.anti_sym`: `pd.Series` with shape `(self.Ne,)`
          with 1 at the free and east half-edges and -1
          at the opposite half-edges.

        Notes
        -----

        - East and west is resepctive to some orientation at the
          moment the indices are computed the partition stays valid as
          long as there are no changes in the topology, so due to vertex
          displacement, 'east' and 'west' might not stay valid. This is
          just a practical naming convention.

        - As the name suggest, this method is not working for edges in
          3D pointing *exactly* north or south, ie iff `edge['dx'] ==
          edge['dy'] == 0`. Until we need or find a better solution,
          we'll just assert it worked.
        """

        if "opposite" not in self.edge_df.columns:
            self.edge_df["opposite"] = get_opposite(self.edge_df)

        self.dble_edges = self.edge_df[self.edge_df["opposite"] >= 0].index
        theta = np.arctan2(
            self.edge_df.loc[self.dble_edges, "dy"],
            self.edge_df.loc[self.dble_edges, "dx"],
        )

        self.east_edges = self.edge_df.loc[self.dble_edges][
            (theta > 0) & (theta <= np.pi)
        ].index
        self.west_edges = pd.Index(
            self.edge_df.loc[self.east_edges, "opposite"].astype(np.int), name="edge"
        )

        self.free_edges = self.edge_df[self.edge_df["opposite"] == -1].index
        self.sgle_edges = self.free_edges.append(self.east_edges)
        self.srtd_edges = self.sgle_edges.append(self.west_edges)

        # Index over the east and free edges, then the opposite indexed
        # by their east counterpart
        self.wrpd_edges = self.sgle_edges.append(self.east_edges)

        self.Ni = self.east_edges.size  # number of inside (east) edges
        self.Nd = self.dble_edges.size  # number of non free half edges
        self.No = self.free_edges.size  # number of free halfedges
        try:
            assert (2 * self.Ni + self.No) == self.Ne
            assert self.west_edges.size == self.Ni
            assert self.Nd == 2 * self.Ni
        # - BC -#
        # Not sure how to build
        # input data so the partition
        # fails (so we can see
        # if the exception is
        # correctly raised).
        # Leaving it in the coverage
        # anyway.
        except AssertionError:
            raise AssertionError(
                """
            Inconsistent partition:
            total half-edges: %s
            number of free: %s
            number of east: %s
            number of west: %s"""
                % (self.Ne, self.No, self.Ni, self.west_edges.size)
            )

        # Anti symetric vector (1 at east and free edges, -1 at opposite)
        self.anti_sym = pd.Series(np.ones(self.Ne), index=self.edge_df.index)
        self.anti_sym.loc[self.west_edges] = -1

    def sort_edges_eastwest(self):
        """reorder edges such the free edges are first,
        then the first half of the double edges, then the other half of
        the double edges, this way, each subset of the edges dataframe
        are contiguous.
        """
        self.get_extra_indices()
        self.edge_df = self.edge_df.loc[self.srtd_edges]
        self.reset_index()
        self.reset_topo()
        self.get_extra_indices()

    def extract(self, face_mask, coords=["x", "y", "z"]):
        """ Extract a new sheet from the sheet
        that correspond to a key word that define a face.

        Parameters
        ----------

        face_mask : column name in face composed by boolean value
        coords :

        Returns
        -------
        sheet_fold_patch_extract :
            subsheet corresponding to the fold patch area.

        """

        datasets = {}
        mask = self.face_df[face_mask].astype(bool)
        datasets["face"] = self.face_df[mask].copy()
        datasets["edge"] = self.edge_df[
            self.edge_df["face"].isin(datasets["face"].index)
        ].copy()
        datasets["vert"] = self.vert_df.loc[self.edge_df["srce"].unique()].copy()

        subsheet = Sheet("subsheet", datasets, self.specs)
        subsheet.reset_index()
        subsheet.reset_topo()
        return subsheet

    def extract_bounding_box(
        self, x_boundary=None, y_boundary=None, z_boundary=None, coords=["x", "y", "z"]
    ):
        """Extracts a new sheet from the embryo sheet

        that correspond to boundary coordinate define by the user.

        Parameters
        ----------
        x_boundary : pair of floats
        y_boundary : pair of floats
        z_boundary : pair of floats
        coords : list of strings, default ['x', 'y', 'z']
          coordinates over which to crop the sheet

        Returns
        -------
        subsheet : a new :class:`Sheet` object

        """
        x, y, z = coords
        datasets = {}
        datasets["face"] = self.face_df.copy()

        if x_boundary is not None:
            xmin, xmax = x_boundary
            datasets["face"] = datasets["face"][
                (datasets["face"][x] > xmin) & (datasets["face"][x] < xmax)
            ].copy()

        if y_boundary is not None:
            ymin, ymax = y_boundary
            datasets["face"] = datasets["face"][
                (datasets["face"][y] > ymin) & (datasets["face"][y] < ymax)
            ].copy()

        if z_boundary is not None:
            zmin, zmax = z_boundary
            datasets["face"] = datasets["face"][
                (datasets["face"][z] > zmin) & (datasets["face"][z] < zmax)
            ].copy()

        datasets["edge"] = self.edge_df[
            self.edge_df["face"].isin(datasets["face"].index)
        ].copy()

        datasets["vert"] = self.vert_df.loc[self.edge_df["srce"].unique()].copy()

        subsheet = Sheet("subsheet", datasets, self.specs)
        subsheet.reset_index()
        subsheet.reset_topo()
        return subsheet

    @classmethod
    def planar_sheet_2d(cls, identifier, nx, ny, distx, disty, noise=None):
        """Creates a planar sheet from an hexagonal grid of cells.

        Parameters
        ----------
        identifier : string
        nx, ny : int
          number of cells in the x and y axes
        distx, disty : float,
          the distances in x and y between the cells
        noise : float, default None
          position noise on the hexagonal grid

        Returns
        -------
        planar_sheet: a 2D :class:`Sheet` instance
          in the (x, y) plane

        """
        from scipy.spatial import Voronoi
        from ..config.geometry import planar_spec
        from ..generation import hexa_grid2d, from_2d_voronoi

        grid = hexa_grid2d(nx, ny, distx, disty, noise)
        datasets = from_2d_voronoi(Voronoi(grid))
        return cls(identifier, datasets, specs=planar_spec(), coords=["x", "y"])

    @classmethod
    def planar_sheet_3d(cls, identifier, nx, ny, distx, disty, noise=None):
        """Creates a planar sheet from an hexagonal grid of cells.

        Parameters
        ----------
        identifier : string
        nx, ny : int
          number of cells in the x and y axes
        distx, disty : float,
          the distances in x and y between the cells
        noise : float, default None
          position noise on the hexagonal grid

        Returns
        -------
        flat_sheet: a 2.5D :class:`Sheet` instance
        """

        from scipy.spatial import Voronoi
        from ..config.geometry import flat_sheet
        from ..generation import hexa_grid2d, from_2d_voronoi

        grid = hexa_grid2d(nx, ny, distx, disty, noise)
        datasets = from_2d_voronoi(Voronoi(grid))
        datasets["vert"]["z"] = 0
        datasets["face"]["z"] = 0

        return cls(identifier, datasets, specs=flat_sheet(), coords=["x", "y", "z"])


def get_opposite(edge_df):
    """
    Returns the indices opposite to the edges in `edge_df`
    """

    st_indexed = (
        edge_df[["srce", "trgt"]].reset_index().set_index(["srce", "trgt"], drop=False)
    )
    flipped = st_indexed.index.swaplevel(0, 1)
    flipped.names = ["srce", "trgt"]
    try:
        opposite = st_indexed.reindex(flipped)["edge"].values
    # see https://github.com/pandas-dev/pandas/issues/21770, will be fixed in
    # pandas 0.24
    except ValueError:
        dup = flipped.duplicated()
        warnings.warn(
            "Duplicated (`srce`, `trgt`) values in edge_df, maybe sanitize your input"
        )
        opposite = st_indexed[~dup].reindex(flipped)["edge"].values
    opposite[np.isnan(opposite)] = -1
    return opposite.astype(np.int)


def get_outer_sheet(eptm):
    """Return a Sheet object formed by all the faces w/o an opposite
    face.
    """
    eptm.get_opposite_faces()
    is_free_face = eptm.face_df["opposite"] == -1
    is_free_edge = eptm.upcast_face(is_free_face)
    edge_df = eptm.edge_df[is_free_edge].copy()
    face_df = eptm.face_df[is_free_face].copy()
    vert_df = eptm.vert_df.loc[edge_df["srce"].unique()].copy()

    datasets = {"edge": edge_df, "face": face_df, "vert": vert_df}
    specs = {k: eptm.specs.get(k, {}) for k in ["face", "edge", "vert", "settings"]}

    return Sheet(eptm.identifier + "outer", datasets, specs)



import pandas as pd
import numpy as np

from ..config.geometry import planar_spec, bulk_spec, flat_sheet

from .utils import make_df

"""
Generate datasets and epithelia from Voronoi tessalations
-------------------------
"""


def from_3d_voronoi(voro):
    """ Creates 3D (bulk geometry) datasets from a Voronoï  tessalation

    Parameters
    ----------
    voro: a :class:`scipy.spatial.Voronoi` object

    Returns
    -------
    datasets: dict
      datasets suitable for :class:`Epithelium` implementation

    """
    specs3d = bulk_spec()

    el_idx = []

    for f_idx, (rv, rp) in enumerate(zip(voro.ridge_vertices, voro.ridge_points)):

        if -1 in rv:
            continue
        face_verts = voro.vertices[rv]
        f_center = face_verts.mean(axis=0)
        c0 = voro.points[rp[0]]
        ctof = f_center - c0

        for rv0, rv1 in zip(rv, np.roll(rv, 1, axis=0)):
            fv0 = voro.vertices[rv0]
            fv1 = voro.vertices[rv1]
            edge_v = fv1 - fv0
            fto0 = fv0 - f_center
            normal = np.cross(fto0, edge_v)
            dotp = np.dot(ctof, normal)
            if np.sign(dotp) > 0:
                el_idx.append([rv0, rv1, f_idx, rp[0]])
                el_idx.append([rv1, rv0, f_idx, rp[1]])
            else:
                el_idx.append([rv1, rv0, f_idx, rp[0]])
                el_idx.append([rv0, rv1, f_idx, rp[1]])

    el_idx = np.array(el_idx)

    coords = ["x", "y", "z"]
    edge_idx = pd.Index(range(el_idx.shape[0]), name="edge")
    edge_df = make_df(edge_idx, specs3d["edge"])

    for i, elem in enumerate(["srce", "trgt", "face", "cell"]):
        edge_df[elem] = el_idx[:, i]

    vert_idx = pd.Index(range(voro.vertices.shape[0]), name="vert")
    vert_df = make_df(vert_idx, specs3d["vert"])
    vert_df[coords] = voro.vertices
    included_verts = edge_df["srce"].unique()
    vert_df = vert_df.loc[included_verts].copy()

    cell_idx = pd.Index(range(voro.points.shape[0]), name="cell")
    cell_df = make_df(cell_idx, specs3d["cell"])
    cell_df[coords] = voro.points
    included_cells = edge_df["cell"].unique()
    cell_df = cell_df.loc[included_cells].copy()

    nfaces = len(voro.ridge_vertices)
    face_idx = pd.Index(np.arange(nfaces), name="face")
    face_df = make_df(face_idx, specs3d["face"])
    included_faces = edge_df["face"].unique()
    face_df = face_df.loc[included_faces].copy()

    edge_df.sort_values(by="cell", inplace=True)

    datasets = {"vert": vert_df, "edge": edge_df, "face": face_df, "cell": cell_df}
    return datasets


def from_2d_voronoi(voro, specs=None):
    """ Creates 2D (sheet geometry) datasets from a Voronoï  tessalation

    Parameters
    ----------
    voro: a :class:`scipy.spatial.Voronoi` object

    Returns
    -------
    datasets: dict
      datasets suitable for :class:`Epithelium` implementation

    """
    if specs is None:
        specs = planar_spec()
    el_idx = []

    for rv, rp in zip(voro.ridge_vertices, voro.ridge_points):

        if -1 in rv:
            continue
        f_center = voro.points[rp[0]]
        for rv0, rv1 in zip(rv, np.roll(rv, 1, axis=0)):
            fv0 = voro.vertices[rv0]
            fv1 = voro.vertices[rv1]
            edge_v = fv1 - fv0
            fto0 = fv0 - f_center
            normal = np.cross(fto0, edge_v)
            if np.sign(normal) > 0:
                el_idx.append([rv0, rv1, rp[0]])
            else:
                el_idx.append([rv0, rv1, rp[1]])

    el_idx = np.array(el_idx)
    coords = ["x", "y"]
    edge_idx = pd.Index(range(el_idx.shape[0]), name="edge")
    edge_df = make_df(edge_idx, specs["edge"])

    for i, elem in enumerate(["srce", "trgt", "face"]):
        edge_df[elem] = el_idx[:, i]

    vert_idx = pd.Index(range(voro.vertices.shape[0]), name="vert")
    vert_df = make_df(vert_idx, specs["vert"])

    vert_df[coords] = voro.vertices

    face_idx = pd.Index(range(voro.points.shape[0]), name="face")
    face_df = make_df(face_idx, specs["face"])
    face_df[coords] = voro.points

    datasets = {"vert": vert_df, "edge": edge_df, "face": face_df}
    return datasets



import warnings
import numpy as np
import logging
import pandas as pd

logger = logging.getLogger(name=__name__)


def _to_2d(df):
    df_2d = to_nd(df, 2)
    return df_2d


def _to_3d(df):
    df_3d = to_nd(df, 3)
    return df_3d


def to_nd(df, ndim):
    """
    Give a new shape to an input data by duplicating its column.

    Parameters
    ----------

    df : input data that will be reshape
    ndim : dimension of the new reshape data.

    Returns
    -------

    df_nd : return array reshaped in ndim.

    """
    df_nd = np.asarray(df).repeat(ndim).reshape((df.size, ndim))
    return df_nd


def combine_specs(*specs):

    combined = {}
    for spec in specs:
        for key in spec:
            if key in combined:
                combined[key].update(spec[key])
            else:
                combined[key] = spec[key]
    return combined


def spec_updater(specs, new):
    """
    Add element to the new dictionary to the specs dictionary.
    Update value if the key already exist.

    Parameters
    ----------
    specs: specification that will be modified
    new: dictionary of new specification
    """
    for key, spec in specs.items():
        if new.get(key) is not None:
            spec.update(new[key])


def set_data_columns(datasets, specs, reset=False):
    """Sets the columns of the dataframes in the datasets dictionnary to
    the uniform values in the specs sub-dictionnaries.

    Parameters
    ----------
    datasets : dict of dataframes
    specs : dict of dicts
    reset : bool, default False

    For each key in specs, the value is a dictionnary whose
    keys are column names for the corresponding dataframe in
    datasets. If there is no such column in the dataframe,
    it is created. If the columns allready exists and  reset is `True`,
    the new value is used.
    """

    for name, spec in specs.items():
        if not len(spec):
            continue
        if "setting" in name:
            continue
        df = datasets.get(name)
        if df is None:
            warnings.warn(
                f"There is no {name} dataset, so the {name}" " spec have no effect."
            )
            continue
        for col, default in spec.items():
            if col in df.columns and reset:
                logger.warning(
                    "Reseting column %s of the %s" " dataset with new specs", col, name
                )
            if col not in df.columns or reset:
                df[col] = default


def data_at_opposite(sheet, edge_data, free_value=None):
    """
    Returns a pd.DataFrame with the values of the input edge_data
    at the opposite edges. For free edges, optionaly replaces Nan values
    with free_value

    Parameters
    ----------
    sheet: a :class:`Sheet` instance
    edge_data:  dataframe contain value of edge

    Returns
    -------
    opposite: pandas series contain value of opposite edge
    """
    if isinstance(edge_data, pd.Series):
        opposite = pd.Series(
            edge_data.reindex(sheet.edge_df["opposite"]).values, index=edge_data.index
        )
    else:
        opposite = pd.DataFrame(
            edge_data.reindex(sheet.edge_df["opposite"]).values, index=edge_data.index
        )
    if free_value is not None:
        opposite = opposite.replace(np.nan, free_value)

    return opposite


def get_sub_eptm(eptm, edges, copy=False):
    """
    Define sub-epithelium corresponding to the edges.

    Parameters
    ----------
    eptm: a :class:`Epithelium` instance
    edges: list of edges includes in the sub-epithelium

    Returns
    -------
    sub_eptm: a :class:`Epithelium` instance
    """
    from ..core.objects import Epithelium

    datasets = {}
    edge_df = eptm.edge_df.loc[edges]
    if edge_df.empty:
        warnings.warn("Sub epithelium appears to be empty")
        return None
    datasets["edge"] = edge_df
    datasets["vert"] = eptm.vert_df.loc[set(edge_df["srce"])]
    datasets["face"] = eptm.face_df.loc[set(edge_df["face"])]
    if "cell" in eptm.datasets:
        datasets["cell"] = eptm.cell_df.loc[set(edge_df["cell"])]

    if copy:
        for elem, df in datasets.items():
            datasets[elem] = df.copy()

    sub_eptm = Epithelium("sub", datasets, eptm.specs)
    sub_eptm.datasets["edge"]["edge_o"] = edges
    sub_eptm.datasets["edge"]["srce_o"] = edge_df["srce"]
    sub_eptm.datasets["edge"]["trgt_o"] = edge_df["trgt"]
    sub_eptm.datasets["edge"]["face_o"] = edge_df["face"]
    if "cell" in eptm.datasets:
        sub_eptm.datasets["edge"]["cell_o"] = edge_df["cell"]

    sub_eptm.datasets["vert"]["srce_o"] = set(edge_df["srce"])
    sub_eptm.datasets["face"]["face_o"] = set(edge_df["face"])
    if "cell" in eptm.datasets:
        sub_eptm.datasets["cell"]["cell_o"] = set(edge_df["cell"])

    sub_eptm.reset_index()
    sub_eptm.reset_topo()
    return sub_eptm


def single_cell(eptm, cell, copy=False):
    """
    Define epithelium instance for all element to a define cell.

    Parameters
    ----------
    eptm : a :class:`Epithelium` instance
    cell : identifier of a cell
    copy : bool, default `False`

    Returns
    -------
    sub_etpm: class:'Epithelium' instance corresponding to the cell
    """
    edges = eptm.edge_df[eptm.edge_df["cell"] == cell].index
    return get_sub_eptm(eptm, edges, copy)


def scaled_unscaled(func, scale, eptm, geom, args=(), kwargs={}, coords=None):
    """Scales the epithelium by an homotetic factor `scale`, applies
    the function `func`, and scales back to original size.

    Parameters
    ----------
    func: the function to apply to the scaled epithelium
    scale: float, the scale to apply
    eptm: a :class:`Epithelium` instance
    geom: a :class:`Geometry` class
    args: sequence, the arguments to pass to func
    kwargs: dictionary, the keywords arguments
      to pass to func
    coords: the coordinates on which the scaling applies

    If the execution of function fails, the scaling is still reverted

    Returns
    -------
    res: the result of the function func
    """
    if coords is None:
        coords = eptm.coords
    geom.scale(eptm, scale, coords)
    geom.update_all(eptm)
    try:
        res = func(*args, **kwargs)
    except:
        raise
    finally:
        geom.scale(eptm, 1 / scale, coords)
        geom.update_all(eptm)
    return res


def modify_segments(eptm, modifiers):
    """Modifies the datasets of a segmented epithelium
    according to the passed modifiers.

    Parameters
    ----------
    eptm : :class:`tyssue.Epithelium`
    modifiers : nested dictionnary

    Note
    ----
    This functions assumes that the epithelium has a `segment_index`
    method as implemented in the :class:`tyssue.Monolayer`.

    Example
    -------
    >>> modifiers = {
    >>>     'apical' : {
    >>>         'edge': {'line_tension': 1.},
    >>>         'face': {'prefered_area': 0.2},
    >>>     },
    >>>     'basal' : {
    >>>         'edge': {'line_tension': 3.},
    >>>         'face': {'prefered_area': 0.1},
    >>>     }
    >>> modify_segments(monolayer, modifiers)
    >>> monolayer.ver_df.loc[monolayer.apical_edges,
    >>>                      'line_tension'].unique()[0] == 1.
    True
    """

    for segment, spec in modifiers.items():
        for element, parameters in spec.items():
            idx = eptm.segment_index(segment, element)
            for param_name, param_value in parameters.items():
                eptm.datasets[element].loc[idx, param_name] = param_value


def _compute_ar(df, coords):
    u, v = coords
    major = np.ptp(df[u].values)
    minor = np.ptp(df[v].values)
    if major < minor:
        minor, major = major, minor
    return 0 if minor == 0 else major / minor


def ar_calculation(sheet, coords=["x", "y"]):
    """ Calculates the aspect ratio of each face of the sheet

    Parameters
    ----------
    eptm : a :class:`Sheet` object
    coords : list of str, optional, default ['x', 'y']
      the coordinates on which to compute the aspect ratio

    Returns
    -------
    AR: pandas series of aspect ratio for all faces.

    Note
    ----
    As is the case in ImageJ, the returned aspect ratio is always higher than 1

    """
    srce_pos = sheet.upcast_srce(sheet.vert_df[sheet.coords])
    srce_pos["face"] = sheet.edge_df["face"]
    return srce_pos.groupby("face").apply(_compute_ar, coords)



"""The `generation` module provides utilities to easily create :class:`Epithelium` objects.
"""

from .utils import *
from .hexagonal_grids import *
from .from_voronoi import *
from .modifiers import *
from .shapes import *
